# Movable Type (r) (C) 2001-2020 Six Apart Ltd. All Rights Reserved.
# This code cannot be redistributed without permission from www.sixapart.com.
# For more information, consult your Movable Type license.
#
# $Id$

package MT::FileSynchronizer::FTPBase;
use strict;

use MT::FileMgr;
use MT::FileSynchronizer;
@MT::FileSynchronizer::FTPBase::ISA = qw( MT::FileSynchronizer );

use File::Find;
use Net::FTP;
use Net::FTP::Find::Mixin;
use MT::Util qw( format_ts offset_time_list );
use MT::SyncSetting;
use Net::Cmd;

sub MAX_RETRY_COUNT () {5}

sub init {
    my $fsyncer = shift;
    $fsyncer->SUPER::init(@_);

    # MT::FileMgr::Local
    $fsyncer->{local} = MT::FileMgr->new('Local');

    $fsyncer;
}

sub synchronize {
    my $fsyncer         = shift;
    my $args            = shift;
    my $remote_dir      = $args->{remote_dir};
    my $local_dir       = $args->{local_dir};
    my $blog            = $args->{blog};
    my $sync_setting_id = $args->{setting_id};
    my $ts              = $args->{ts};
    my $full_sync       = $args->{full_sync};
    my $avoid_dot_ht    = $args->{avoid_dot_ht};

    my $ftp  = $fsyncer->{ftp};
    my $fmgr = $fsyncer->{local};

    my $pack = MT->component("Sync");

    require MT::Util::Log;
    MT::Util::Log::init();

    require POSIX;
    my @now      = offset_time_list( time, $blog );
    my $start_ts = POSIX::strftime( "%Y%m%d%H%M00", @now );

    my $site_url = $blog->site_url;
    $site_url =~ s!(^http(s|)://[^/]*).*!$1!;

    my $tmp_dir = MT::SyncSetting::get_temp_dir_path($sync_setting_id);
    unless ( -d $tmp_dir ) {
        $fmgr->mkpath($tmp_dir);
    }
    my $count_file = File::Spec->catdir( $tmp_dir, 'count_file_' . $ts );

    my $remote_base_dir = $ftp->pwd();
    MT::Util::Log->info( " Remote base Dir:" . $remote_base_dir );
    MT::Util::Log->info( " Remote Dir:" . $remote_dir );

    my $full_remote_dir
        = $fsyncer->_get_realpath( $remote_base_dir, $remote_dir );

    MT::Util::Log->info( " Full Remote Dir:" . $full_remote_dir );

    unless ($full_remote_dir) {
        return $fsyncer->sync_error(
            $pack->translate(
                "Cannot access to remote directory '[_1]'", $remote_dir
            ),
            $count_file,
            $blog,
            $sync_setting_id,
        );
    }

    $ftp->{ftp}->cwd($remote_base_dir);

    my $count_data
        = $fmgr->exists($count_file) ? $fmgr->get_data($count_file) : '';
    my $task_number
        = defined $count_data ? ( split ',', $count_data )[1] || '0' : '0';
    my $count = 0;

    my $tmp_file = &_catdir( $tmp_dir, 'tmp_file' );

    require MT::SyncFileList;
    my @sync_file_list = MT::SyncFileList->load(
        { blog_id => $blog->id, sync_setting_id => $sync_setting_id } );
    my %last_local_file_list;

    MT::Util::Log->info(' Start analyze SyncFileList.');

    foreach my $sync_file (@sync_file_list) {
        $count++;
        MT::Util::Log->info( '  process,' . $task_number . ',' . $count );
        $fmgr->put_data( 'process,' . $task_number . ',' . $count,
            $count_file );
        return $fsyncer->sync_error(
            $pack->translate(
                "Error writing to \'[_1]\': [_2]", $count_file,
                $fmgr->errstr,
            ),
            $count_file,
            $blog,
            $sync_setting_id,
        ) if ( defined $fmgr->errstr );

        my $path = &_catdir( $local_dir, $sync_file->path );
        if ( $fmgr->exists($path) ) {
            my %params = (
                id     => $sync_file->id,
                path   => $sync_file->path,
                is_dir => $sync_file->is_dir,
                ts     => $sync_file->timestamp,
            );
            $last_local_file_list{ $sync_file->path } = \%params;
        }
        else {
            my $path        = $sync_file->path;
            my $fullpath    = &_catdir( $full_remote_dir, $path );
            my $remote_path = $fsyncer->_encode_str($fullpath);
            if (   !$full_sync
                && !$sync_file->is_dir
                && $ftp->exists($remote_path) )
            {
                # Remove files
                $ftp->delete($remote_path)
                    or return $fsyncer->sync_error(
                    $pack->translate(
                        "Deleting file '[_1]' failed.", $fullpath
                    ),
                    $count_file,
                    $blog,
                    $sync_setting_id,
                    );
                MT::Util::Log->info( '  Delete: ' . $remote_path );
            }
            my %params = (
                id     => $sync_file->id,
                path   => $remote_path,
                is_dir => $sync_file->is_dir,
            );
            push @{ $fsyncer->{delete_file_list} }, \%params;
        }
    }

    MT::Util::Log->info(' End   analyze SyncFileList.');

    # Remove empty directories
    unless ($full_sync) {
        my @delete_file_list = ();
        @delete_file_list
            = sort { $b->{path} cmp $a->{path} }
            @{ $fsyncer->{delete_file_list} }
            if defined $fsyncer->{delete_file_list};

        MT::Util::Log->info(' Start Remove empty directories.');

        foreach my $delete_file (@delete_file_list) {
            my $remote_path = $delete_file->{path};
            my $ftp         = $fsyncer->{ftp}->{ftp};
            if ( $delete_file->{is_dir} && $ftp->cwd($remote_path) ) {
                my $count = 0;
                $ftp->finddepth(
                    {   wanted => sub {
                            return if $_ eq '.';
                            $count++;
                        },
                        use_mlsd => 1,
                    },
                    $remote_path
                );
                next if $count;
                $ftp->rmdir($remote_path)
                    or return $fsyncer->sync_error(
                    $pack->translate(
                        "Deleting path '[_1]' failed.", $remote_path
                    ),
                    $count_file,
                    $blog,
                    $sync_setting_id,
                    );
                MT::Util::Log->info( '  Delete Dir: ' . $remote_path );
            }
        }
    }

    MT::Util::Log->info(' End   Remove empty directories.');

    my @local_file_list;

    MT::Util::Log->info(' Start scan local files.');

    if ( MT->is_callback_registered('scan_local_file_list.FTP') ) {
        MT->run_callbacks(
            'scan_local_file_list.FTP',
            $fsyncer,
            {   blog_id      => $blog->id,
                setting_id   => $sync_setting_id,
                full_sync    => $full_sync,
                local_dir    => $local_dir,
                avoid_dot_ht => $avoid_dot_ht,
                fmgr         => $fmgr,
                args         => $args,
            },
            \@local_file_list,
        );
    }
    else {
        @local_file_list
            = @{ scan_local_file_list( $local_dir, $avoid_dot_ht ) };
    }

    MT::Util::Log->info(' End   scan local files.');

    my @invalid_filenames = map { $_->{path} }
        grep { $_->{path} =~ /.*[\.|\s]$/; } @local_file_list;
    return $fsyncer->sync_error(
        $pack->translate(
            'Directory or file by which end of name is dot(.) or blank exists. Cannot synchronize these files.: "[_1]"',
            join( '", "', @invalid_filenames ),
        ),
        $count_file,
        $blog,
        $sync_setting_id,
    ) if @invalid_filenames;

    my @local_paths = map { $_->{path} } @local_file_list;
    MT->run_callbacks( 'content_delivery_file_filter',
        $blog->id, $sync_setting_id, \@local_paths );
    my %allowed_local_paths = map { $_ => 1 } @local_paths;

    my @err_files;
    my $too_many_errors = 0;

    MT::Util::Log->info(' Start analyze local files.');

    foreach my $data (@local_file_list) {
        my $path = $data->{path};
        my $enc  = $data->{enc};

        next unless $allowed_local_paths{$path};

        $count++;
        $fmgr->put_data( 'process,' . $task_number . ',' . $count,
            $count_file );
        return $fsyncer->sync_error(
            $pack->translate(
                "Error writing to \'[_1]\': [_2]", $count_file,
                $fmgr->errstr,
            ),
            $count_file,
            $blog,
            $sync_setting_id,
        ) if ( defined $fmgr->errstr );

        my $name = $path;
        $name =~ s/^\Q$local_dir//;
        next unless $name;

        my $local_encoded_path = Encode::encode( $enc, $path );

        my $is_dir   = ( -d $local_encoded_path ) ? 1 : 0;
        my $mod_time = $fmgr->file_mod_time($local_encoded_path);
        my %params   = (
            path   => $name,
            is_dir => $is_dir,
            ts     => $mod_time,
        );

        my $last_local_file = $last_local_file_list{$name};
        if ( !$full_sync && $last_local_file ) {
            next if $last_local_file->{is_dir};
            next if $params{ts} <= $last_local_file->{ts};
            $params{id} = $last_local_file->{id};
            push @{ $fsyncer->{update_file_list} }, \%params;
        }
        else {
            $params{id} = $last_local_file->{id} if $last_local_file;
            push @{ $fsyncer->{update_file_list} }, \%params;
            if ( -d $local_encoded_path ) {
                my $raw_path
                    = &_catdir( $remote_base_dir, $remote_dir, $name );
                my $remote_encoded_path = $fsyncer->_encode_str($raw_path);
                $ftp->mkpath($remote_encoded_path);
                next;
            }
        }

        my $upload_source_path;
        my $ext
            = ( File::Basename::fileparse( $local_encoded_path, '\.[^\.]*' ) )
            [2];
        if ( ( !( defined $ext ) || ( defined $ext && lc($ext) ne '.pdf' ) )
            && -T $local_encoded_path )
        {
            my $file = $fmgr->get_data( $local_encoded_path, 'upload' );
            return $fsyncer->sync_error( $fmgr->errstr, $count_file, $blog,
                $sync_setting_id )
                if $fmgr->errstr;
            if ($site_url) {
                $file = $fsyncer->replace_url( $file, $site_url );
            }
            $fmgr->put_data( $file, $tmp_file, 'upload' );
            if ( defined $fmgr->errstr ) {
                return $fsyncer->sync_error(
                    $pack->translate(
                        "Unable to write temporary file ([_1]): [_2]",
                        $tmp_file, $fmgr->errstr
                    ),
                    $count_file,
                    $blog,
                    $sync_setting_id,
                );
            }
            MT::Util::Log->info( '  put temporary file (' . $tmp_file . ')' );

            $upload_source_path = $tmp_file;
        }
        else {
            $upload_source_path = $local_encoded_path;
        }

        my $local_size = $fmgr->file_size($upload_source_path) || 0;
        return $fsyncer->sync_error(
            $pack->translate(
                "Unable to get size of temporary file ([_1]): [_2]",
                $tmp_file, $fmgr->errstr
            ),
            $count_file,
            $blog,
            $sync_setting_id,
        ) if ( defined $fmgr->errstr );

        my $raw_path = &_catdir( $remote_base_dir, $remote_dir, $name );
        my $remote_encoded_path = $fsyncer->_encode_str($raw_path);

        my $remote_size = 0;
        for ( my $i = 0; $i < MAX_RETRY_COUNT; $i++ ) {

            # 1+ Retry
            MT::Util::Log->info("  Retry $i") if $i;

            my $dir
                = $is_dir ? $name : ( File::Basename::fileparse $name )[1];
            my $encoded_dir       = $fsyncer->_encode_str($dir);
            my $base_path         = &_catdir( $remote_base_dir, $remote_dir );
            my $encoded_base_path = $fsyncer->_encode_str($base_path);
            my $encoded_dest = &_catdir( $encoded_base_path, $encoded_dir );
            unless ( $ftp->{ftp}->cwd($encoded_dest) ) {
                my @path = split '/', $encoded_dir;
                my $real_dest
                    = $fsyncer->_mk_remote_dir( $encoded_base_path, @path );
                unless ($real_dest) {
                    $ftp->{ftp}->quit;
                    my $type
                        = ref($ftp) eq 'MT::FileMgr::FTP' ? 'FTP' : 'FTPS';
                    undef $ftp;
                    $ftp
                        = MT::FileMgr->new( $type, $fsyncer->{ftp_host},
                        $fsyncer->{ftp_id}, $fsyncer->{ftp_password},
                        $fsyncer->{ftp_port} )
                        or return $fsyncer->sync_error(
                        $pack->translate(
                            "FTP reconnection was failed. ([_1])", $@
                        ),
                        $count_file,
                        $blog,
                        $sync_setting_id,
                        );
                    return $fsyncer->sync_error(
                        $pack->translate( "FTP connection lost.", ),
                        $count_file, $blog, $sync_setting_id, )
                        unless $ftp->list();
                }
            }

            MT::Util::Log->info(' Start put data.');
            $remote_size = eval {
                $ftp->put( $upload_source_path, $remote_encoded_path,
                    'upload' );
            } || 0;
            if ( $@ or $remote_size !~ /^\A[0-9]+\z/ ) {
                my $err = $@ || $remote_size;
                if ( $err =~ /Timeout/ ) {
                    $err = $pack->translate("FTP connection timeout.");
                }
                MT::Util::Log->error( $err,
                    $count_file, $blog, $sync_setting_id );
            }
            else {
                MT::Util::Log->info( '  put data:'
                        . "'$upload_source_path, $remote_encoded_path'" );
                MT::Util::Log->info('  End   put data.');

                # put ok
                last
                    if ( $ftp->last_status() == CMD_OK
                    && $local_size == $remote_size );
            }
            MT::Util::Log->info( '  last status: ' . $ftp->last_status() );

            MT::Util::Log->error(
                $pack->translate(
                    "Unable to write remote files ([_1]): [_2]",
                    "'$upload_source_path,$remote_encoded_path",
                    $ftp->errstr
                )
            );
        }
        if (   $ftp->last_status() != CMD_OK
            || $local_size != $remote_size )
        {
            MT::Util::Log->error(
                $pack->translate(
                    "  tried [_1] times but failed. ([_2]): [_3]",
                    MAX_RETRY_COUNT, "'$raw_path'", $ftp->errstr
                )
            );
        }

        if ( defined $ftp->errstr || $local_size != $remote_size ) {
            $raw_path = '' unless defined $raw_path;
            push @err_files, "'$raw_path,$local_size,$remote_size'";
            if ( @err_files >= 20 ) {
                $too_many_errors = 1;
                last;
            }
        }
    }
    if (@err_files) {
        my $files = join ', ', @err_files;
        my $errstr
            = $too_many_errors
            ? "Too many errors"
            : "Unable to write remote files";
        return $fsyncer->sync_error(
            $pack->translate(
                $errstr
                    . ". Please check activity log for more details.: [_1]",
                $ftp->errstr
            ),
            $count_file,
            $blog,
            $sync_setting_id,
            $pack->translate(
                $errstr . " ([_1]): [_2]",
                $files, $ftp->errstr
            ),
        );
    }

    MT::Util::Log->info(' End   analyze local files.');

    return $start_ts;
}

sub scan_local_file_list {
    my $local_dir    = shift;
    my $avoid_dot_ht = shift;
    my $default_enc  = $^O ne 'MSWin32' ? 'utf8' : 'cp932';
    my @local_file_list;
    File::Find::find(
        {   wanted => sub {
                my $path = $File::Find::name;
                require File::Basename;
                my $filename = File::Basename::basename($path);
                return
                       if -d $path
                    || $filename eq "mt-config.cgi"
                    || $filename =~ m/^mt-preview-.+/;

                # Are .htaccess and .htpasswd synchronized?
                return
                    if $avoid_dot_ht
                    && ( $filename eq ".htaccess" || $filename eq ".htpasswd" );

                my $enc  = MT::I18N::guess_encoding($path) || $default_enc;
                my $decoded_path = Encode::decode( $enc, $path );
                push @local_file_list, { path => $decoded_path, enc => $enc };
            },
            no_chdir => 1,
        },
        $local_dir
    );
    return \@local_file_list;
}

sub _get_realpath {
    my $fsyncer = shift;
    my ( $base_path, $path ) = @_;
    my $ftp = $fsyncer->{ftp};

    my @path      = split '/', $path;
    my $real_path = $base_path;
    $real_path = $fsyncer->_mk_remote_dir( $real_path, @path );
    $ftp->{ftp}->cwd($base_path);
    return $real_path;
}

sub _catdir {
    my @list = @_;
    return MT::Util::caturl(@list);
}

sub _mk_remote_dir {
    my $fsyncer = shift;
    my ( $real_path, @path ) = @_;

    my $ftp = $fsyncer->{ftp};

    for ( my $i = 0; $i <= $#path; $i++ ) {
        $real_path = &_catdir( $real_path, $path[$i] );
        if ( $ftp->{ftp}->cwd($real_path) ) {
            $real_path = $ftp->{ftp}->pwd()
                or return;
        }
        else {
            # If there is no sync directory, will be created.
            $ftp->mkpath($real_path)
                or return;
            $ftp->{ftp}->cwd($real_path)
                or return;
            $real_path = $ftp->{ftp}->pwd()
                or return;
        }
    }
    return $real_path;
}

sub _encode_str {
    my $fsyncer = shift;
    my $str     = shift;

    my $encoded_str
        = $fsyncer->{ostype} eq 'unix'    ? Encode::encode_utf8($str)
        : $fsyncer->{ostype} eq 'windows' ? Encode::encode( 'cp932', $str )
        :                                   $_;

    return $encoded_str;
}

1;
